# Miscellaneous utilities.

export issomething, 
        order_asc, 
        isdisjoint, 
        pushrand!, 
        get_bit, 
        init_array, 
        subseteq_fast, 
        similar!,
        Var,
        Lit, 
        var2lit, 
        lit2var, 
        variables,
        num_variables, 
        always, 
        never, 
        uniform, 
        logsumexp, 
        noop


using CUDA: CuArray, CuVector, CuMatrix, CUDA

"Is the argument not `nothing`?"
@inline issomething(x) = !isnothing(x)

"Order the arguments in a tuple in ascending order"
@inline order_asc(x, y) = x > y ? (y, x) : (x , y)

"Are the given sets disjoint (no shared elements)?"
function Base.isdisjoint(set1::AbstractSet, set2::AbstractSet, set3::AbstractSet, othersets::AbstractSet...)::Bool
    seen = set1 # can be sped up by copying seen first then reusing it with union!?
    for set in [set2, set3, othersets...]
        if !isempty(intersect(seen,set))
            return false
        else
            seen = union(seen, set)
        end
    end
    return true
end

"""
Push `element` into random position in vector`v`
"""
function pushrand!(v::AbstractVector{<:Any}, element)
    len = length(v)
    i = rand(1:len + 1)
    if i == len + 1
        push!(v, element)
    else
        splice!(v, i:i, [element, v[i]])
    end
    v
end

"Retrieve the jth bit from a `BitVector` chunk"
get_bit(chunk, j) = chunk & (UInt64(1) << Base._mod64(j-1)) != 0

"An array of undetermined values (fast) for the given element type"
@inline init_array(::Type{T}, dims::Int...) where T<:Number = Array{T}(undef, dims...)
@inline init_array(::Type{T}, dims::Int...) where T<:Bool= BitArray(undef, dims...)

"Replacement for BitSet.⊆ that does not allocate a new BitSet"
function subseteq_fast(s::BitSet,t::BitSet) # see https://discourse.julialang.org/t/issubset-bitset-bitset-is-slow-and-allocates-unnecessarily/43828
    # TODO see if https://github.com/JuliaLang/julia/pull/36882/files has fixed this issue in 1.6
    s === t && return true
    s.offset < t.offset && return false
    length(s.bits) > length(t.bits) && return false
    length(s) > length(t) && return false
    return all(e -> e ∈ t, s)
end

"Reuse a given array if it has the right type and size, otherwise make a new one"
function similar!(reuse, ::Type{A}, desired_size...) where A<:AbstractArray
    if reuse isa A && size(reuse) == (desired_size...,)
        reuse
    elseif reuse isa A && ndims(reuse) == 1 && length(desired_size) == 1
        # Use resize! to save some effort when the desired array is 1-dimentional
        resize!(reuse, desired_size...)
    else
        A(undef, desired_size...)
    end
end

#####################
# General logic
#####################

"""
Variables are represented as 32-bit unsigned integers
"""
const Var = UInt32 # variable ids

"""
Literals are represented as 32-bit signed integers.
Positive literals are positive integers identical to their variable. Negative literals are their negations. Integer 0 should not be used to represent literals.
"""
const Lit = Int32 # variable with a positive or negative sign

"Convert a variable to the corresponding positive literal"
@inline var2lit(v::Var)::Lit = convert(Lit,v)

"Convert a literal its variable, removing the sign of the literal"
@inline lit2var(l::Lit)::Var = convert(Var,abs(l))

"Get the `BitSet` of variables in the data structure"
function variables end

"Number of variables in the data structure"
@inline num_variables(x)::Int = length(variables(x))

#####################
# probability semantics
#####################

"An array of 100% probabilities for the given element type"
@inline always(::Type{T}, dims::Int...) where T<:Number = ones(T, dims...)
@inline always(::Type{T}, dims::Int...) where T<:Bool = trues(dims...)
@inline always(::Type{<:CuArray{T}}, dims::Int...) where T = CUDA.ones(T, dims...)
@inline always(::Type{T}, dims::Int...) where T<:AbstractArray = always(eltype(T), dims...)::T

"An array of 0% probabilities for the given element type"
@inline never(::Type{T}, dims::Int...) where T<:Number = zeros(T, dims...)
@inline never(::Type{T}, dims::Int...) where T<:Bool = falses(dims...)
@inline never(::Type{<:CuArray{T}}, dims::Int...) where T = CUDA.zeros(T, dims...)
@inline never(::Type{T}, dims::Int...) where T<:AbstractArray = never(eltype(T), dims...)::T

"An array of uniform probabilities"
@inline uniform(dims::Int...) = uniform(Float64, dims...)
@inline uniform(::Type{F}, dims::Int...) where F<:AbstractFloat = always(F, dims...) ./ prod(dims)

import StatsFuns: logsumexp #extend

"Marginalize out dimensions `dims` from log-probability tensor"
function logsumexp(a::AbstractArray, dims)
    return dropdims(mapslices(logsumexp, a, dims=dims), dims=dims)
end

#####################
# functional programming
#####################

"Function that does nothing"
@inline noop(_...) = nothing
